/* 
 * extend.c --
 *
 *      Contains extensions to the basic TCL command set.
 *
 */

/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "tcl.h"

/*
 *----------------------------------------------------------------------
 *
 * Tcl_EchoCmd --
 *    Implements the TCL echo command:
 *        echo str1 [str2..]
 *    See the extend(TCL) manual page.
 *
 * Results:
 *      Always returns TCL_OK.
 *
 *----------------------------------------------------------------------
 */
int
Tcl_EchoCmd(clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
int i;

for (i = 1;i < argc ; i++) 
    {
    fputs(argv[i], stdout);
    if (i < (argc-1))
        printf(" ");
    }
printf("\n");
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_IncrCmd --
 *     Implements the TCL incr command:
 *         incr var [value]
 *     See the extend(TCL) manual page.
 *
 * Results:
 *      Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
Tcl_IncrCmd (clientdata, interp, argc, argv)
    ClientData  clientdata;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
long   varval, incval;
char *varptr;

if ((argc < 2) || (argc > 3))
    return Tcl_RetError (interp, "TCL", "bad # arg: %s variable value", 
                         argv [0]);

varptr = Tcl_GetVar(interp, argv[1], 0);
if (varptr == NULL)
    return Tcl_RetError (interp, "TCL", "variable or array element not found (%s)", argv [1]);

if (!StrToLong (varptr, 10, &varval))
    return Tcl_RetError (interp, "TCL", "non-numeric variable contents for %s",
                         argv[1]);

if (argc == 3)
    {
    if (!StrToLong (argv[2], 10, &incval))
        return Tcl_RetError (interp, "TCL", "non-numeric increment value",
                             argv[2]);
    }
else
    incval = 1;

sprintf (interp->result,"%ld", varval + incval);
Tcl_SetVar (interp, argv[1], interp->result, 0);
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_SplitCmd --
 *     Implements the TCL split command:
 *         split splitchars string
 *      See the extend(TCL) manual page.
 *
 * Results:
 *      Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
Tcl_SplitCmd (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        int    nsplitchars;
        int    strlength;
        int    i, j, matches;
        char  *s, *chars;
        char **fragments;
        int    frag;
        char  *result;

        if (argc != 3)
                return Tcl_RetError (interp, "TCL", 
                                "bad # arg: %.50s splitchars string", argv[0]);

        chars = argv[1];
        s = argv[2];
        strlength = strlen(s);
        nsplitchars = strlen(chars);

        for (i = 0, matches = 0; i < strlength; i++)
        {
                for (j = 0; j < nsplitchars; j++)
                {
                        if (s[i] == chars[j])
                        {
                                matches++;
                                goto nextchar;
                        }
                }
                nextchar: ;
        }

        fragments = (char **)ckalloc(sizeof(char *) * (matches + 2));
        fragments[0] = s;
        if (matches != 0)
        {
                for (i = 0, frag = 1; i < strlength; i++)
                {
                        for (j = 0; j < nsplitchars; j++)
                        {
                                if (s[i] == chars[j])
                                {
                                        s[i] = '\0';
                                        fragments[frag++] = &s[i+1];
                                        goto nextone;
                                }
                        }
                        nextone: ;
                }
        }
        result = Tcl_Merge(matches + 1, fragments);
        ckfree(fragments);
        Tcl_Return(interp, result, TCL_DYNAMIC);
        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_JoinCmd --
 *     Implements the TCL join command:
 *         join joinchar list
 *     See the extend(TCL) manual page.
 *
 * Results:
 *      Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
Tcl_JoinCmd (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        int    strlength;
        int    i, j;
        char  *result;
        int    myargc;
        char **myargv;
        int    is_a_join_char;

        if ((argc != 3) || ((is_a_join_char = strlen(argv[1])) > 1))
                return Tcl_RetError (interp,"TCL", 
                                     "bad # arg: %.50s joinchar list",
                                     argv[0]);

        if (Tcl_SplitList(interp, argv[2], &myargc, &myargv) == TCL_ERROR)
                return TCL_ERROR;

        if (myargc == 0) return TCL_OK;

        for (i = 0, strlength = 0; i < myargc; i++)
        {
                strlength += strlen(myargv[i]) + 1;
        }
        result = ckalloc(strlength + 1);  /* plus room for null byte */
        *result = '\0';
        strcat(result, myargv[0]);
        for (i = 1; i < myargc; i++)
        {
                /* not the ultimate high-performance way to do things... */
                if (is_a_join_char)
                        strcat(result, argv[1]);
                strcat(result, myargv[i]);
        }
        ckfree((char *)myargv);

        Tcl_Return(interp, result, TCL_DYNAMIC);
        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_LoopCmd --
 *     Implements the TCL loop command:
 *         loop var start end [increment] command
 *     See the extend(TCL) manual page.
 *
 * Results:
 *      Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
Tcl_LoopCmd (dummy, interp, argc, argv)
    ClientData  dummy;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    int   result = TCL_OK;
    long  i, lo, hi, incr = 1;
    char *command;

    if ((argc < 5) || (argc > 6))
        return Tcl_RetError (interp, "TCL", 
                        "bad # args: %.50s var lo hi [incr] command", argv[0]);

    if (!StrToLong (argv[2], 10, &lo))
        return Tcl_RetError (interp, "TCL", 
                             "non-numeric loop start value (%s)", argv[2]);
    if (!StrToLong (argv[3], 10, &hi))
        return Tcl_RetError (interp, "TCL", "non-numeric loop end value (%s)", 
                             argv[2]);
    if (argc == 5)
        command = argv[4];
    else
        {
        if (!StrToLong (argv[4], 10, &incr))
            return Tcl_RetError (interp, "TCL", 
                             "non-numeric loop increment value (%s)", argv[2]);
        command = argv[5];
        }
    for (i = lo; (((i < hi) && (incr > 0)) || ((i > hi) && (incr < 0)));
         i += incr) 
        {
        char itxt[12];
        sprintf(itxt,"%ld",i);
        Tcl_SetVar(interp, argv[1], itxt, 0);
        result = Tcl_Eval(interp, command, 0, (char **) NULL);
        if (result != TCL_OK) {
            if (result == TCL_CONTINUE) {
                result = TCL_OK;
            } else if (result == TCL_BREAK) {
                result = TCL_OK;
                break;
            } else if (result == TCL_ERROR) {
                char msg[100];
                sprintf(msg, " (\"loop\" body line %d)", interp->errorLine);
                Tcl_AddErrorInfo(interp, msg);
                break;
            } else {
                break;
            }
        }
    }
    if (result == TCL_OK) 
        *interp->result = 0;
    return result;
}
/*
 *----------------------------------------------------------------------
 *
 * Tcl_ReplaceCmd --
 *     Implements the TCL replace command:
 *         replace var index [string]
 *     See the extend(TCL) manual page.
 *
 * Results:
 *      Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
Tcl_ReplaceCmd (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    int    myargc;
    char **myargv;
    char  *varcontents;
    int    val;
    int    i;
    char  *result;

    if ((argc < 3) || (argc > 4))
        return Tcl_RetError (interp, "TCL", 
                             "bad # arg: %.50s var index [string]", argv[0]);

    varcontents = Tcl_GetVar (interp,argv[1],0);
    if (varcontents == NULL)
        return Tcl_RetError (interp, "TCL", "nonexistent variable or array element (%s)", 
                             argv[1]);

    if (Tcl_SplitList (interp, varcontents, &myargc, &myargv) == TCL_ERROR)
        return TCL_ERROR;

    if (!StrToInt (argv[2], 10, &val))
        return Tcl_RetError (interp, "TCL", "non-numeric index (%s)", argv[2]);

    if ((val < 0) || (myargc <= val))
        return Tcl_RetError (interp, "TCL", "index out of range");

    Tcl_Return(interp, myargv[val], TCL_VOLATILE);

    if (argc == 4)
        myargv[val] = argv[3];
    else
        {
        for (i = val; i < myargc - 1; i++)
            myargv[i] = myargv[i+1];
        myargc = myargc - 1;
        }

    result = Tcl_Merge(myargc, myargv);
    Tcl_SetVar(interp, argv[1], result, 0);
    ckfree(result);
    ckfree(myargv);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_WhileCmd --
 *
 *      This procedure is invoked to process the "while" Tcl command.
 *      See the user documentation for details on what it does.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      See the user documentation.
 *
 *----------------------------------------------------------------------
 */
int
Tcl_WhileCmd(dummy, interp, argc, argv)
    ClientData  dummy;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    int result;
    long value;

    if (argc != 3) 
        return Tcl_RetError (interp, "TCL", "bad # args: %.50s test command", 
                             argv[0]);

    while (1) {
        result = Tcl_Expr(interp, argv[1], &value);
        if (result != TCL_OK)
            return result;

        if (!value)
            break;

        result = Tcl_Eval(interp, argv[2], 0, (char **) NULL);
        if (result == TCL_CONTINUE) {
            result = TCL_OK;
        } else if (result != TCL_OK) {
            if (result == TCL_ERROR) {
                char msg[60];
                sprintf(msg, " (\"while\" body line %d)", interp->errorLine);
                Tcl_AddErrorInfo(interp, msg);
            }
            break;
        }
    }
    if (result == TCL_BREAK) {
        result = TCL_OK;
    }
    if (result == TCL_OK) {
        *interp->result = '\0';
    }
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * sortComparisonRoutine  --
 *      Support routine to pass strcmp to qsort.  Will ignore leading
 *      "{" and white spaces in an element.
 *      
 *
 *----------------------------------------------------------------------
 */
static int 
sortComparisonRoutine (arg1, arg2)
    char **arg1;
    char **arg2;
{
register char *scan1 = *arg1;
register char *scan2 = *arg2;

for (; (*scan1 != '\0') && ((*scan1 == '{') || (isspace (*scan1))); scan1++)
    /* Empty loop body */;
for (; (*scan2 != '\0') && ((*scan2 == '{') || (isspace (*scan2))); scan2++)
    /* Empty loop body */;

return (strcmp(scan1, scan2));
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_SortlistCmd --
 *     Implements the TCL sortlist command:
 *         sortlist list
 *     See the extend(TCL) manual page.
 *
 * Results:
 *    Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
Tcl_SortlistCmd (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
int    myargc;
char **myargv;

if (argc != 2)
    return Tcl_RetError (interp, "TCL", "bad # arg: %.50s list", argv[0]);

if (Tcl_SplitList(interp, argv[1], &myargc, &myargv) == TCL_ERROR)
    return TCL_ERROR;

qsort (myargv, (unsigned)myargc, sizeof(char *), sortComparisonRoutine);

Tcl_Return(interp, Tcl_Merge(myargc, myargv), TCL_DYNAMIC);
ckfree(myargv);
return TCL_OK;
}




