/* 
 * tclUtil2.c --
 *
 *      Contains various utility procedures that have been added to extend
 *      TCL.
 *
 */

/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#include <string.h>
#include <stdio.h>
#include <varargs.h>

#include <ctype.h>
#include <errno.h>

#include <sys/types.h>

#ifdef XENIX
#  define MAXINT 32767
#endif
#ifdef BSD
#  define MAXINT 2147483647
#endif

#if defined(XENIX286) || defined(BSD)
extern int errno;
#endif

#if BSD
#  define TOLOWER tolower
#  define TOUPPER toupper
#else
#  define TOLOWER _tolower
#  define TOUPPER _toupper
#endif

#include "tcl.h"
#include "tclInt.h"

/*
 * Include the error code table for this operating system
 */

#ifdef SYSV_3_2_2
#   include "errSV3_2_2.h"
#endif
#ifdef SYSV_3_2_0
#   include "errSV3_2_0.h"
#endif
#ifdef XENIX286
#   include "errXenix286.h"
#endif
#ifdef XENIX386
#   include "errXenix386.h"
#endif
#ifdef HPUX
#   include "errHPUX.h"
#endif
#ifdef BSD
#   include "errBSD.h"
#endif

#ifndef SYSV_3_2_2
    extern char *strerror();
#endif


/*
 *----------------------------------------------------------------------
 *
 * StrToLong --
 *      Convert an Ascii string to an long number of the specified base.
 *
 * Parameters:
 *      NumStr - String containing a number.
 *      Base - The base to assume for the number 8, 10 or 16.
 *      NumValP - Place to return the converted number.  Will be 
 *        unchanged if there is an error.
 *
 * Results:
 *      Returns 1 if the string was a valid number, 0 invalid.
 *
 *----------------------------------------------------------------------
 */
int
StrToLong (NumStr, Base, NumValP)
    char *NumStr;
    int   Base;
    long *NumValP;
{ /* StrToLong */
int   Negative;
long  RetValue = 0;
char  Digit;

if (Negative = (*NumStr == '-'))
    NumStr++;
if (*NumStr == '0')
    {
    NumStr++;
    Base = (*NumStr == 'x') ? 16 : 8;
    if (Base == 16)
        NumStr++;
    }

if (Base == 10)
    while ((*NumStr >= '0') && (*NumStr <= '9'))
        {
        RetValue = (RetValue * 10) + (*NumStr - '0');
        NumStr++;
        }
else if (Base == 8)
    while ((*NumStr >= '0') && (*NumStr <= '7'))
        {
        RetValue = (RetValue << 3) + (*NumStr - '0');
        NumStr++;
        }
else if (Base == 16)
    while (TRUE)
        {
        if ((*NumStr >= '0') && (*NumStr <= '9'))
            RetValue = (RetValue << 4) + (*NumStr - '0');
        else if ((*NumStr >= 'a') && (*NumStr <= 'f'))
            RetValue = (RetValue << 4) + 10 + (*NumStr - 'a');
        else if ((*NumStr >= 'A') && (*NumStr <= 'F'))
            RetValue = (RetValue << 4) + 10 + (*NumStr - 'A');
        else
            break;
        NumStr++;
        }
else
    return FALSE;  /* Invalid base */
if (*NumStr != '\0')
    return FALSE;  /* Invalid number */

*NumValP = (Negative) ? -RetValue : RetValue;
return TRUE;

} /* StrToLong */

/*
 *----------------------------------------------------------------------
 *
 * StrToInt --
 *      Convert an Ascii string to an number of the specified base.
 *
 * Parameters:
 *      NumStr - String containing a number.
 *      Base - The base to assume for the number 8, 10 or 16.
 *      NumValP - Place to return the converted number.  Will be 
 *        unchanged if there is an error.
 *
 * Results:
 *      Returns 1 if the string was a valid number, 0 invalid.
 *
 *----------------------------------------------------------------------
 */
int
StrToInt (NumStr, Base, NumValP)
    char *NumStr;
    int   Base;
    int  *NumValP;
{ /* StrToInt */
int  NumOk;
long RetValue;

if (NumOk = StrToLong (NumStr, Base, &RetValue))
    *NumValP = RetValue;

return NumOk;

} /* StrToInt */

/*
 *----------------------------------------------------------------------
 *
 * StrToUInt --
 *      Convert an Ascii string to an unsigned int of the specified base.
 *
 * Parameters:
 *      NumStr - String containing a number.
 *      Base - The base to assume for the number 8, 10 or 16.
 *      NumValP - Place to return the converted number.  Will be 
 *        unchanged if there is an error.
 *
 * Results:
 *      Returns 1 if the string was a valid number, 0 invalid.
 *
 *----------------------------------------------------------------------
 */
int
StrToUInt (NumStr, Base, NumValP)
    char      *NumStr;
    int        Base;
    unsigned  *NumValP;
{ /* StrToUInt */
int  NumOk;
long RetValue;

if (NumOk = (StrToLong (NumStr, Base, &RetValue) && (RetValue >= 0)))
    *NumValP = RetValue;

return NumOk;

} /* StrToUInt */

/*
 *----------------------------------------------------------------------
 *
 * DownShift --
 *     Utility procedure to downshift a string.  It is written in such
 *     a way as that the target string maybe the same as the source string.
 *
 *----------------------------------------------------------------------
 */
void 
DownShift (targetStr, sourceStr)
    char *targetStr;
    char *sourceStr;
{
register char *targetPtr = targetStr;
register char *sourcePtr = sourceStr;
register char  theChar;

for (; (theChar = *sourcePtr) != '\0'; sourcePtr++) 
    {
    if (isupper (theChar))
        theChar = TOLOWER (theChar);
    *targetPtr++ = theChar;
    }
*targetPtr = '\0';
}

/*
 *----------------------------------------------------------------------
 *
 * UpShift --
 *     Utility procedure to upshift a string.  It is written in such
 *     a way as that the target string maybe the same as the source string.
 *
 *----------------------------------------------------------------------
 */
void 
UpShift (targetStr, sourceStr)
    char *targetStr;
    char *sourceStr;
{
register char *targetPtr = targetStr;
register char *sourcePtr = sourceStr;
register char  theChar;

for (; (theChar = *sourcePtr) != '\0'; sourcePtr++)
    {
    if (islower (theChar))
        theChar = TOUPPER (theChar);
    *targetPtr++ = theChar;
    }
*targetPtr = '\0';
}

/*
 *----------------------------------------------------------------------
 *
 * InitStrFile
 *     Utility procedure for BSD to initializea stdio FILE control block
 *     to point to a string to write.  This is a total kludge to interface
 *     to the _doprnt procedure, which is the way to implement the vsprintf
 *     procedure on BSD.  Yuk.  BSD needs a vsprintf.
 *
 *----------------------------------------------------------------------
 */
#ifdef BSD
void
InitStrFile (buffer, strCBPtr)
    char *buffer;
    FILE *strCBPtr;
{
strCBPtr->_cnt = 0;
strCBPtr->_ptr = buffer;
strCBPtr->_base = buffer;
strCBPtr->_bufsiz = MAXINT;
strCBPtr->_flag = _IOWRT;
strCBPtr->_file = 0;

}
#endif

/*
 *----------------------------------------------------------------------
 *
 * GetSysErrCode --
 *     Convert a operating system error number to its symbolic error code.
 *     This looks the specified error number up in a system dependant table.
 *
 * Results:
 *     A pointer to a static string.
 *
 *----------------------------------------------------------------------
 */
char *
GetSysErrCode (errorNum)
    int errorNum;
{
register int idx;

for (idx = 0; (ErrorCodeTab [idx].errNum  != errorNum) && 
              (ErrorCodeTab [idx].errNum  != -1);
     idx++);
return ErrorCodeTab [idx].errCode;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_SetErrorCode --
 *     Format strings into the errorCode global variable.
 *
 * Parameters:
 *      interp - Pointer to current interpreture date structure.
 *      formatStr - Sprintf formating specs to encode the vararg parameters
 *        into the error code.  Result MUST fit into a TCL_RESULT_SIZE buffer.
 *      args - Variable arguments list to sprintf.
 *
 * Side effects:
 *     The errorCode global TCL variable will have its value altered.
 *
 *----------------------------------------------------------------------
 */
void
Tcl_SetErrorCode (interp, formatStr, va_alist)
    Tcl_Interp *interp;
    char       *formatStr;
    va_dcl
{
char     codeBuf [TCL_RESULT_SIZE];
va_list  args;
#ifdef BSD
FILE     strCB;
#endif

/*
 * Place a marker to detect result overflow.
 */

codeBuf [TCL_RESULT_SIZE-1] = '\0';

va_start (args);
#ifndef BSD
vsprintf (codeBuf, formatStr, args);
#else
InitStrFile (codeBuf, &strCB);
_doprnt (formatStr, args, &strCB);
*strCB._ptr = '\0';
#endif
va_end (args);

if (codeBuf [TCL_RESULT_SIZE-1] != '\0')
    panic ("Tcl_SetErrorCode: buffer overflow.");

Tcl_SetVar (interp, "errorCode", codeBuf, 1);

}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_RetError --
 *      Return a result of an error message.
 *      
 * Parameters:
 *      interp - Pointer to current interpreture date structure.
 *      errorCode - The value to set the global variable errorCode to.
 *        This is a list, whose first element is a subsystem ID.  The 
 *        remainding elements of the list are subsystem dependant.  If
 *        NULL is specified, then the error code is not set, it should
 *        have been set previously by a call to Tcl_SetErrorCode.  This
 *        route should be taken when there is a complex error code to
 *        encode.
 *      formatStr - Sprintf formating specs to encode the vararg parameters
 *        into the error message.  Result MUST fit into a TCL_RESULT_SIZE
 *        buffer.
 *      args - Variable arguments list to encode into formatStr.
 *
 * Results:
 *      Always returns TCL_ERROR, thus this function can be called as a
 *      argument to a return from the command procedure.
 *
 *----------------------------------------------------------------------
 */
int
Tcl_RetError (interp, errorCode, formatStr, va_alist)
    Tcl_Interp *interp;
    char       *errorCode;
    char       *formatStr;
    va_dcl
{ /* Tcl_RetError */
Interp  *iPtr = (Interp *) interp;
va_list  args;
#ifdef BSD
FILE     strCB;
#endif

/*
 * Place a marker to detect result overflow, Make sure it points to 
 * resultSpace.
 */

ResetResult (iPtr);
iPtr->result [TCL_RESULT_SIZE-1] = '\0';

va_start (args);
#ifndef BSD
vsprintf (iPtr->result, formatStr, args);
#else
InitStrFile (iPtr->result, &strCB);
_doprnt (formatStr, args, &strCB);
*strCB._ptr = '\0';
#endif
va_end (args);

if (iPtr->result [TCL_RESULT_SIZE-1] != '\0')
    panic ("Tcl_RetError: buffer overflow.");

if (errorCode != NULL)
    Tcl_SetErrorCode (iPtr, "%s", errorCode);

return TCL_ERROR;

} /* Tcl_RetError */

/*
 *----------------------------------------------------------------------
 *
 * Tcl_RetSysError --
 *      Return a result of the system error message for the current value
 *      of errno. 
 *      
 * Parameters:
 *      interp - Pointer to current interpreture date structure.
 *      formatStr - Sprintf formating specs to encode the vararg parameters
 *        into the error message.  Result MUST fit into a TCL_RESULT_SIZE
 *        buffer.
 *      args - Variable arguments list to encode into formatStr.
 *
 * Results:
 *      Always returns TCL_ERROR, thus this function can be called as a
 *      argument to a return from the command procedure.
 *
 * Side effects:
 *      The TCL global variable errorCode is set to contain the subsystem,
 *      "UNIX", an symbolic unix error number and the error message string.
 *
 *----------------------------------------------------------------------
 */
int
Tcl_RetSysError (interp, formatStr, va_alist)
    Tcl_Interp *interp;
    char       *formatStr;
    va_dcl
{ /* Tcl_RetSysError */
Interp  *iPtr = (Interp *) interp;
va_list  args;
char    *sysErrMsg;
char    *sysErrCode;
#ifdef BSD
FILE     strCB;
#endif

/*
 * Do these upfront incase something sets errno
 */

sysErrMsg = strerror (errno);
sysErrCode =GetSysErrCode (errno);

/*
 * Place a marker to detect result overflow, Make sure it points to 
 * resultSpace.
 */

ResetResult (iPtr);
iPtr->result [TCL_RESULT_SIZE-1] = '\0';

*iPtr->result = '\0';
if (formatStr != NULL)
    { 
    va_start (args);
#ifndef BSD
    vsprintf (iPtr->result, formatStr, args);
#else
    InitStrFile (iPtr->result, &strCB);
    _doprnt (formatStr, args, &strCB);
    *strCB._ptr = '\0';
#endif
    va_end (args);
    strcat (iPtr->result, ": ");
    }
strcat (iPtr->result, sysErrMsg);

if (iPtr->result [TCL_RESULT_SIZE-1] != '\0')
    panic ("Tcl_RetSysError: buffer overflow.");

Tcl_SetErrorCode (iPtr, "UNIX %s {%s}", sysErrCode, sysErrMsg);
return TCL_ERROR;

} /* Tcl_RetSysError */
