/* 
 * clock.c --
 *
 *      Contains the TCL time and date related commands.
 *
 */

/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#include <sys/types.h>
#include <time.h>

char *getenv();

#include "tcl.h"

#if !(defined (BSD) || defined(XENIX286))
#    define HAVE_TZSET
#endif

/*
 * Struct used to hold timezone information for each interpreter.
 */

typedef struct timeZoneInfo_t {
            int   useLocal;
            char *localTZstr;
            char *GmtTZstr;
            } timeZoneInfo_t;
typedef struct timeZoneInfo_t *timeZoneInfo_pt;

/*
 *----------------------------------------------------------------------
 *
 * CmdGetclock --
 *     Implements the TCL getclock command:
 *     See the clock(TCL) manual page.
 *
 * Results:
 *     Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CmdGetclock (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
if (argc != 1)
    return Tcl_RetError (interp, "TCL", "bad # arg: %.50s", argv[0]);

sprintf (interp->result, "%ld", time ((VOID *)NULL));
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdFmtclock --
 *     Implements the TCL fmtclock command:
 *         fmtclock clockval [format]
 *     See the clock(TCL) manual page.
 *
 * Results:
 *     Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CmdFmtclock (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
timeZoneInfo_pt  TZInfoP = (timeZoneInfo_pt) clientData;
long             ClockVal;
char            *format;
struct tm       *timeDataPtr;
int              fmtError;
extern struct tm *localtime(), *gmtime();

if ((argc < 2) || (argc > 3))
    return Tcl_RetError (interp, "TCL", "bad # arg: %.50s clockval [format]", 
                         argv[0]);

if (!StrToLong (argv[1], 10, &ClockVal) || (ClockVal < 0))
    return Tcl_RetError (interp, "TCL", "Invalid clock value", argv[0]);

if (argc == 3)
    format = argv[2];
else
    format = "%a %b %d %H:%M:%S %Z %Y";

if (TZInfoP->useLocal)
    timeDataPtr = localtime (&ClockVal);
else    
    timeDataPtr = gmtime (&ClockVal);

#if defined(SYSV_3_2_2) || defined(HPUX)
fmtError = strftime (interp->result, TCL_RESULT_SIZE, format, timeDataPtr) < 0;
#else
fmtError = ascftime (interp->result, format, timeDataPtr) < 0;
#endif
if (fmtError)
    return Tcl_RetError (interp, "TCL", "Invalid parameter to %s", argv[0]);

return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdSettimezone --
 *     Implements the TCL settimezone command:
 *         settimezone tzspec
 *     See the clock(TCL) manual page.
 *
 * Results:
 *     Standard TCL results, may return the UNIX system error message.
 *
 * Side effects:
 *     Changes the TZ environment variable, this affects all interperters
 *     running in the current process, even though the timezone user by 
 *     fmtclock is interperter local.  This also affects all subprocesses 
 *     run by any interperter in the current process.
 *
 *----------------------------------------------------------------------
 */
int
CmdSettimezone (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
timeZoneInfo_pt  TZInfoP = (timeZoneInfo_pt) clientData;

#ifndef HAVE_TZSET
return Tcl_RetError (interp, "TCL", 
                    "%s is not available in this version of Unix", argv[0]);
#else
if (argc != 2)
    return Tcl_RetError (interp, "TCL", "bad # arg: %.50s tzspec", argv[0]);

if (strcmp (argv[1], "local") == 0)
    TZInfoP->useLocal = TRUE;
else if (strcmp (argv[1], "GMT") == 0)
    TZInfoP->useLocal = FALSE;
else
    return Tcl_RetError (interp, "TCL", 
                         "expected \"local\" or \"GMT\" for timezone");
if (TZInfoP->useLocal)
    putenv (TZInfoP->localTZstr);
else
    putenv (TZInfoP->GmtTZstr);
tzset();

return TCL_OK;
#endif
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_InitClock --
 *   Initialize clock commands and install them into Tcl
 *
 *----------------------------------------------------------------------
 */
Tcl_InitClock (interp)
    Tcl_Interp *interp;
{
timeZoneInfo_pt  TZInfoP = NULL;
char            *TZenvPtr;

TZInfoP = (timeZoneInfo_pt)ckalloc (sizeof (timeZoneInfo_t));
TZInfoP->useLocal = TRUE;

#ifdef HAVE_TZSET
TZenvPtr = getenv ("TZ");
TZInfoP->localTZstr = ckalloc (strlen (TZenvPtr) + 4);
strcpy (TZInfoP->localTZstr, "TZ=");
strcat (TZInfoP->localTZstr, TZenvPtr);

TZInfoP->GmtTZstr = "TZ=GMT";
#endif


Tcl_CreateCommand (interp, "getclock", CmdGetclock, (ClientData)NULL, 
                  (void (*)())NULL);
Tcl_CreateCommand (interp, "fmtclock", CmdFmtclock, (ClientData)TZInfoP, 
                  (void (*)())NULL);
Tcl_CreateCommand (interp, "settimezone", CmdSettimezone, (ClientData)TZInfoP, 
                  (void (*)())NULL);
}
