/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#ifndef NOSCCSID
        static char *sccs_id = "%M% %I% %H%";
#endif

#include "tcl.h"
#include <sys/types.h>

#ifdef XENIX
#  define MAXINT 32767
#endif
#ifdef BSD
#  define MAXINT 2147483647
#endif

#if !(defined (XENIX) || defined(BSD))
#  include <values.h>
#endif

#define MININT (-MAXINT)-1

extern int rand();


/*
 *----------------------------------------------------------------------
 *
 * CmdMax --
 *      Implements the TCL max command:
 *        max num1 num2 [..numN]
 *     See the math(TCL) manual page.
 *
 * Results:
 *      Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CmdMax (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
int MaxVal = MININT;
int MaxIdx = 1;
int Value, idx;
        

if (argc < 3)
        return Tcl_RetError (interp, "TCL", 
                             "bad # arg: %.50s num1 num2 [..numN]", argv[0]);

for (idx = 1; idx < argc; idx++)
        {
        if (!StrToInt (argv[idx], 10, &Value))
                return Tcl_RetError (interp, "TCL", 
                                     "invalid numeric argument");
        if (Value > MaxVal)
                {
                MaxVal = Value;
                MaxIdx = idx;
                }
        }
strcpy (interp->result, argv[MaxIdx]);
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdMin --
 *     Implements the TCL min command:
 *         min num1 num2 [..numN]
 *     See the math(TCL) manual page.
 *
 * Results:
 *      Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CmdMin (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
int MinVal = MAXINT;
int MinIdx = 1;
int Value, idx;
        

if (argc < 3)
        return Tcl_RetError (interp, "TCL", 
                             "bad # arg: %.50s num1 num2 [..numN]", argv[0]);

for (idx = 1; idx < argc; idx++)
        {
        if (!StrToInt (argv[idx], 10, &Value))
                return Tcl_RetError (interp, "TCL", 
                                     "invalid numeric argument");
        if (Value < MinVal)
                {
                MinVal = Value;
                MinIdx = idx;
                }
        }
strcpy (interp->result, argv[MinIdx]);
return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * really_random --
 *     Insure a good random return for a range, unlike an arbitrary
 *     random() % n, thanks to Ken Arnold, Unix Review, October 1987.
 *
 *----------------------------------------------------------------------
 */
#ifdef BSD
#    define RANDOM_RANGE ((1 << 31) - 1)
#else
#    define RANDOM_RANGE ((1 << 15) - 1)
#endif
static int 

really_random (my_range)
    int my_range;
{
        int max_multiple, rnum;

        max_multiple = RANDOM_RANGE / my_range;
        max_multiple *= my_range;
        while ((rnum = rand()) >= max_multiple)
                continue;
        return(rnum % my_range);
}

/*
 *----------------------------------------------------------------------
 *
 * CmdRandom  --
 *     Implements the TCL random command:
 *         random limit
 *     See the math(TCL) manual page.
 *
 * Results:
 *      Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CmdRandom (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        int range;

        if (argc != 2)
                return Tcl_RetError (interp, "TCL", "bad # arg: %.50s limit", 
                                     argv[0]);

        if ((!StrToInt (argv[1], 10, &range)) ||
            (range <= 0))
                return Tcl_RetError (interp, "TCL", "range must be > 0: %.50s",
                                     argv[0]);

        if (range > RANDOM_RANGE)
            return Tcl_RetError (interp, "TCL", "range must be <= %d: %.50s",
				 RANDOM_RANGE, argv[0]);

        sprintf(interp->result, "%d", really_random(range));
        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdSeed --
 *     Implements the TCL seed command:
 *         seed [seedval]
 *     See the math(TCL) manual page.
 *
 * Results:
 *      Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CmdSeed (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        int seed;

        if (argc > 2)
                return Tcl_RetError (interp, "TCL", 
                                     "bad # arg: %.50s [seedval]", argv[0]);

        if (argc == 2)
                {
                if (!StrToInt (argv[1], 10, &seed))
                        return Tcl_RetError (interp, "TCL", 
                                             "non-numeric seed");
                }
        else
                seed = (unsigned)(getpid() + time((time_t *)NULL));

        srand(seed);
        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 *  Tcl_InitMath --
 *      Initialize the TCL file scanning facility..
 *
 *----------------------------------------------------------------------
 */
Tcl_InitMath (interp)
    Tcl_Interp *interp;
{
Tcl_CreateCommand (interp, "max", CmdMax, (ClientData)NULL, 
                  (void (*)())NULL);
Tcl_CreateCommand (interp, "min", CmdMin, (ClientData)NULL, 
                  (void (*)())NULL);
Tcl_CreateCommand (interp, "random", CmdRandom, (ClientData)NULL, 
                  (void (*)())NULL);
Tcl_CreateCommand (interp, "seed", CmdSeed, (ClientData)NULL,
                  (void (*)())NULL);
}


