/* 
 * permission.c --
 *
 *      Contains the symbolic UNIX file permission parser for the chmod
 *      command.
 *
 */

/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#include "tcl.h"

/*
 *----------------------------------------------------------------------
 *
 * ConvSymMode --
 *      Parse and convert symbolic file permissions as specified by chmod(C).
 *
 * Parameters:
 *      SymPermsP - The symbolic permissions to parse.
 *      ModeVal - The existing permissions value on a file.
 *
 * Results:
 *      The new permissions, or -1 if invalid permissions where supplied.
 *
 *----------------------------------------------------------------------
 */
int
ConvSymMode (SysPermsP, ModeVal)
  char *SysPermsP;
  int   ModeVal;

{ /* ConvSymMode */
int  User, Group, Other;
char Operator;
int  RWX_Mask, SetUID, Sticky, Locking;
int  NewMode;

while (*SysPermsP != 0)
  { /* Scan and decode symbolic mode specification */
  User = Group = Other = FALSE;

  while (! ((*SysPermsP == '+') || (*SysPermsP == '-') || (*SysPermsP == '=')))
    { /* Scan who field */
    switch (*SysPermsP) {
      case 'a':
        User = Group = Other = TRUE;
        break;
      case 'u':
        User = TRUE;
        break;
      case 'g':
        Group = TRUE;
        break;
      case 'o':
        Other = TRUE;
        break;
      default:
        return -1;  /* Error, not a recognized character or EOL */
      } /* switch */
    SysPermsP++;
    } /* Scan who field */

  /* If none where specified, that means all */

  if (! (User || Group || Other))
    User = Group = Other = TRUE;

  Operator = *SysPermsP++;

  /* Decode the permissions */

  RWX_Mask = 0;
  SetUID = Sticky = Locking = FALSE;

  while (! ((*SysPermsP == ',') || (*SysPermsP == 0)))
    { /* Scan permissions field */
    switch (*SysPermsP) {
      case 'r':
        RWX_Mask |= 4;
        break;
      case 'w':
        RWX_Mask |= 2;
        break;
      case 'x':
        RWX_Mask |= 1;
        break;
      case 's':
        SetUID = TRUE;
        break;
      case 't':
        Sticky = TRUE;
        break;
      case 'l':
        Locking = TRUE;
        break;
      default:
        return -1;  /* Error, not a recognized character */
      } /* switch */
    SysPermsP++;
    } /* Scan permissions field */

  /* Build mode map of specified values */

  NewMode = 0;
  if (User)
    NewMode |= RWX_Mask << 6;
  if (Group)
    NewMode |= RWX_Mask << 3;
  if (Other)
    NewMode |= RWX_Mask;
  if (SetUID && User)
    NewMode |= 04000;
  if ((SetUID || Locking) && Group)
    NewMode |= 02000;
  if (Sticky)
    NewMode |= 01000;

  /* Add to cumulative mode based on operator */

  if (Operator == '+')
    ModeVal |= NewMode;
  else if (Operator == '-')
    ModeVal &= ~NewMode;
  else if (Operator == '=')
    ModeVal = NewMode;
  if (*SysPermsP == ',')
    SysPermsP++;
  } /* Scan and decode symbolic mode specification */

return ModeVal;

} /* ConvSymMode */

/*
 *----------------------------------------------------------------------
 *
 * ConvAbsMode --
 *      Convert a absolute octal file mode.  If this fails, it might be a 
 *      symbolic mode.
 *
 * Parameters:
 *      ModeStrP - The file mode string.
 *
 * Results:
 *      The new permissions, or -1 if invalid permissions where supplied.
 *
 *----------------------------------------------------------------------
 */
int
ConvAbsMode (ModeStrP)
  char *ModeStrP;

{ /* ConvAbsMode */
int   Mode;

if (!StrToInt (ModeStrP, 8, &Mode))
  return -1;  /* Not an octal number */ 
else
  return Mode;

} /* ConvAbsMode */
