/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#include "tcl.h"

#ifndef NOSCCSID
    static char *sccs_id = "%M% %I% %H%";
#endif

#include "regexp.h"

char *boyer_moore_search_compile();
char *boyer_moore_search_execute();


/*
 *----------------------------------------------------------------------
 *
 * InitRegExpInfo --
 *     Initialize regular expression information structure.
 *
 *----------------------------------------------------------------------
 */
void
InitRegExpInfo (regExpPtr)
    regexp_pt regExpPtr;
{
    regExpPtr->plainText = NULL;
    regExpPtr->progPtr = NULL;
    regExpPtr->BoyerMoorePtr = NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * CleanRegExpInfo --
 *     Free all resources associated with a regular expression info 
 *     structure..
 *
 *----------------------------------------------------------------------
 */
void
CleanRegExpInfo (regExpPtr)
    regexp_pt regExpPtr;
{
    if (regExpPtr->plainText != NULL)
    	ckfree (regExpPtr->plainText);
    if (regExpPtr->progPtr != NULL)
    	ckfree (regExpPtr->progPtr);
    if (regExpPtr->BoyerMoorePtr != NULL)
    	ckfree (regExpPtr->BoyerMoorePtr);
    InitRegExpInfo (regExpPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * FindNonRegExpSubStr
 *     Find the largest substring that does not have any regular 
 *     expression meta-characters and is not located within `[...]'.
 *
 *----------------------------------------------------------------------
 */
void
FindNonRegExpSubStr (expression, subStrPtrP, subStrLenP)
    char  *expression;
    char **subStrPtrP;
    int   *subStrLenP;
{
    register char *subStrPtr = NULL;
    register char  subStrLen = 0;
    register char *scanPtr   = expression;
    register int   len;

    while (*scanPtr != '\0') {
        len = strcspn (scanPtr, META);
        /*
         * If we are at a meta-character, by-pass till non-meta.  If we hit
         * a `[' then by-pass the entire `[...]' range, but be careful, could
         * have omitted `]'.
         */
        if (len == 0) {
            scanPtr += strspn (scanPtr, META_NO_BRACKET);
            if (*scanPtr == '[') {
                scanPtr += strcspn (scanPtr, "]");
                if (*scanPtr == ']')
                    scanPtr++;
            }          
        } else {
            if (len > subStrLen) {
                subStrPtr = scanPtr;
                subStrLen = len;
            }
            scanPtr += len;
        }
    }
    *subStrPtrP = subStrPtr;
    *subStrLenP = subStrLen;
}

/*
 *----------------------------------------------------------------------
 *
 * RegExpCompile --
 *     Compile a regular expression.
 *
 * Parameters:
 *     regExpP - Used to hold info on this regular expression.
 *     expression - Regular expression to compile.
 *     flags - The following flags are recognized:
 *         NO_CASE - Comparison will be regardless of case.
 *         CACHE_EXP - Do caching of expressions, don't recompile if
 *             expression is the same as the previous.       
 *         BOTH_ALGORITHMS - If specified, a Boyer-Moore expression is 
 *             compiled for the largest substring of the expression that
 *             does not contain any meta-characters.  This is slows 
 *             compiling, but speeds up large searches.
 *
 * Results:
 *     Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
RegExpCompile (interp, regExpP, expression, flags)
    Tcl_Interp  *interp;
    regexp_pt    regExpP;
    char        *expression;
    int          flags;
{
    char *expBuf;
    int   anyMeta;

    if (*expression == '\0')
        return Tcl_RetError (interp, "TCL", "Null regular expression");    

    if (flags & NO_CASE) {
        expBuf = ckalloc (strlen (expression) + 1);
        DownShift (expBuf, expression);
    } else
        expBuf = expression;
    /*
     * If caching is being done check if this matches the previous expression 
     */

    if ((flags & CACHE_EXP) && 
        (regExpP->plainText != NULL) &&
        (((flags & NO_CASE) != 0) == regExpP->noCase) &&
        (*expBuf == *regExpP->plainText) &&
        (strcmp (expBuf, regExpP->plainText) == 0)) {
        /*
         * Same as last expression, don't compile.
         */
        if (flags & NO_CASE)
            ckfree (expBuf);
        return TCL_OK;
    }

    CleanRegExpInfo (regExpP);
    regExpP->noCase = flags & NO_CASE;
    anyMeta = strpbrk (expBuf, META) != NULL;

    /*
     * If no meta-characters, use Boyer-Moore string matching only.
     */
    if (!anyMeta) {
        regExpP->BoyerMoorePtr = 
            boyer_moore_search_compile (expBuf, strlen (expBuf));
        goto OkExitPoint;
    }
 
    /*
     * Build a Boyer-Moore on the largest non-meta substring, if requested.
     */
    if (flags & BOTH_ALGORITHMS) {
        char *subStrPtr;
        int   subStrLen;
        
        FindNonRegExpSubStr (expBuf, &subStrPtr, &subStrLen);
        if (subStrLen > 0)
            regExpP->BoyerMoorePtr = 
                boyer_moore_search_compile (subStrPtr, subStrLen);
    }
    
    /*
     * Compile meta-character containing regular expression.
     */
    if (!RegComp (expBuf, regExpP)) {
        if (flags & NO_CASE)
            ckfree (expBuf);
        CleanRegExpInfo (regExpP);
        return Tcl_RetError (interp, "TCL", "error in regular expression: %s", 
                             regExpP->errMsg);
    }
  
    /* 
     * Cache expression if requested and return all ok.
     */
    OkExitPoint: 
    
    if (flags & CACHE_EXP) {
        if (flags & NO_CASE)
           regExpP->plainText = expBuf;
        else {
           regExpP->plainText = ckalloc(strlen(expression) + 1);
           strcpy (regExpP->plainText, expression);
        }
        regExpP->noCase = (flags & NO_CASE);
    } else
        if (flags & NO_CASE)
            ckfree (expBuf);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * RegExpExecute --
 *     Execute a regular expression compiled with Boyer-Moore and/or 
 *     regexp.
 *
 * Parameters:
 *     regExpP - Used to hold info on this regular expression.
 *     matchStrIn - String to match against the regular expression.
 *     matchStrLower - Optional lower case version of the string.  If
 *       multiple no case matches are being done, time can be saved by
 *       down shifting the string in advance.  NULL if not a no-case 
 *       match or this procedure is to do the down shifting.
 *
 * Results:
 *     TRUE if a match, FALSE if it does not match.
 *
 *----------------------------------------------------------------------
 */
int
RegExpExecute (interp, regExpP, matchStrIn, matchStrLower)
    Tcl_Interp  *interp;
    regexp_pt    regExpP;
    char        *matchStrIn;
    char        *matchStrLower;
{
    char *matchStr;
    int   result;

    if (regExpP->noCase) {
        if (matchStrLower == NULL) {
            matchStr = ckalloc (strlen (matchStrIn) + 1);
            DownShift (matchStr, matchStrIn);
        } else
            matchStr = matchStrLower;
    } else
        matchStr = matchStrIn;
    /*
     * If a Boyer-Moore pattern has been compiled, use that algorithm
     * to test against the text.
     */
    if (regExpP->BoyerMoorePtr != NULL) {
        char     *startPtr;
        unsigned  matchLen;
        startPtr = boyer_moore_search_execute (matchStr, strlen (matchStr), 
                                            regExpP->BoyerMoorePtr, &matchLen);
        /*
         * If we did match, return the range that was matched if not regular
         * expression program is available.  If its is available, fall through
         * and do a regular expression compare.
         */
        if (startPtr == NULL) {
            result = FALSE;
            goto exitPoint;
        } else {
            if (regExpP->progPtr == NULL) {
                regExpP->subExpIdx [0] = startPtr - matchStr;
                regExpP->subExpLen [0] = matchLen;
                result = TRUE;
                goto exitPoint;
            }
        }
    }
    
    /*
     * Give it a go with full regular expressions
     */
    result = RegExec(regExpP, matchStr);

    /*
     * Clean up and return status here.
     */
exitPoint:
    if ((regExpP->noCase) && (matchStrLower == NULL))
        ckfree (matchStr);
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * RegexpCmd --
 *     Implements the TCL regexp command:
 *         regexp string [nocase] [matchvar]
 *              [submatchvar1..submatchvar9]
 *     See the regexp(TCL) manual page.
 *
 * Results:
 *     Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
RegexpCmd (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    regexp_pt       regExpP = (regexp_pt) clientData;
    int             flags = CACHE_EXP;
    int             result;
    
    if ((argc < 3) || (argc > NSUBEXP + 4))
        return Tcl_RetError (interp, "TCL", 
                   "bad # arg: %.50s reg_exp string [nocase] [indexvar]%s",
                   argv[0], " [matchvar1..matchvar9]");
    
    if ((argc >= 4) && (*argv[3] == 'n') && (strcmp (argv[3], "nocase") == 0))
        flags |= NO_CASE;
    
    if (RegExpCompile (interp, regExpP, argv [1], flags) != TCL_OK)
        return TCL_ERROR;

    result = RegExpExecute (interp, regExpP, argv[2], (char *)NULL);

    if (result)
        { /* matched */
        int  argIdx;
        int  strInfoIdx;
        char indexBuf [20];
        /*
         * The encode the match start and length of the match and submatches
         * in each variable supplied.  Null string for a variable name means
         * to not return the value.
         */
        for (argIdx = (flags & NO_CASE) ? 4 : 3, strInfoIdx = 0; 
            argIdx < argc; argIdx++, strInfoIdx++)
        if (*argv[argIdx] != '\0')
            {
            sprintf (indexBuf, "%d %d", regExpP->subExpIdx[strInfoIdx],
                                        regExpP->subExpLen[strInfoIdx]);
            Tcl_SetVar (interp, argv[argIdx], indexBuf, FALSE);
            }

        } /* matched! */

    sprintf(interp->result, "%d", result);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * RegsubCmd --
 *     Implements the TCL Regsub command:
 *         regsub exp string subspec retvar [nocase]
 *     See the regexp(TCL) manual page.
 *
 * Results:
 *     Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
RegsubCmd (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    regexp_pt       regExpP = (regexp_pt) clientData;
    int             flags = CACHE_EXP;
    char           *subResultPtr;
    int             result;
    
    if ((argc < 5) || (argc > 6))
        return Tcl_RetError (interp, "TCL", 
                   "bad # arg: %.50s exp string subspec retvar [nocase]",
                   argv[0]);
    
    if (argc == 6) {
        if (strcmp (argv[5], "nocase") != 0)
            return Tcl_RetError (interp, "TCL", 
                                 "invalid option: nocase expected");
        flags |= NO_CASE;
    }
    if (RegExpCompile (interp, regExpP, argv [1], flags) != TCL_OK)
        return TCL_ERROR;

    result = RegExpExecute (interp, regExpP, argv[2], (char *)NULL);
    if (result)
        { /* matched */
        if ((subResultPtr = RegSub (regExpP, argv[2], argv[3])) == NULL)
            return Tcl_RetError (interp, "TCL", "%s: %s", argv[0], 
                                 regExpP->errMsg);
        
        Tcl_SetVar (interp, argv[4], subResultPtr, FALSE);
        ckfree (subResultPtr);
        } /* matched! */

    sprintf(interp->result, "%d", result);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_InitRegExp --
 *   Initialize regexp command and installs it into Tcl.
 *
 *----------------------------------------------------------------------
 */
Tcl_InitRegExp(interp)
    Tcl_Interp *interp;
{
regexp_pt regExpP;

regExpP = (regexp_pt)ckalloc (sizeof (struct regexp_t));
CleanRegExpInfo (regExpP);

Tcl_CreateCommand (interp, "regexp", RegexpCmd, 
                  (ClientData)regExpP, (void (*)())NULL);
Tcl_CreateCommand (interp, "regsub", RegsubCmd, 
                   (ClientData)regExpP, (void (*)())NULL);
}
