/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#ifndef NOSCCSID
        static char *sccs_id = "%M% %I% %H%";
#endif

#include <stdio.h>

#include "tcl.h"

/*
 * If these sizes change, some printfs must also change.  Sorry.
 */
#define ARG_TRUNCATE_SIZE 40
#define CMD_TRUNCATE_SIZE 60

struct TraceInfo_t {
    Tcl_Trace TraceHolder;
    int       NoEval;
    int       NoTruncate;
    int       Depth;
    };
typedef struct TraceInfo_t *TraceInfo_pt;


/*
 *----------------------------------------------------------------------
 *
 * PrintArg --
 *     Print an argument string, truncating and adding "..." if its longer
 *     then ARG_TRUNCATE_SIZE.  If the string contains white spaces, quote
 *     it with angle brackets.
 *
 *----------------------------------------------------------------------
 */
static void
PrintArg (ArgStr, NoTruncate)
    char *ArgStr;
    int   NoTruncate;
{
        int i;
        int quote_it = FALSE;
        char *p;

        for (i = 0, p = ArgStr; ((i < ARG_TRUNCATE_SIZE) && (*p != '\0'));
             i++, p++)
                if (*p == ' ')
                        {
                        quote_it = TRUE;
                        break;
                        }

        if (*ArgStr == '\0') 
                quote_it = TRUE;

        if (quote_it) putchar('<');

        if (NoTruncate || (strlen(ArgStr) < ARG_TRUNCATE_SIZE))
                printf("%s", ArgStr);
        else
                printf("%.40s...", ArgStr);

        if (quote_it) 
                putchar('>');
        putchar(' ');
}

/*
 *----------------------------------------------------------------------
 *
 * TraceRoutine --
 *      Routine called by Tcl_Eval to trace a command.
 *
 *----------------------------------------------------------------------
 */
void
TraceRoutine (TraceInfoP, interp, level, command, cmdProc, cmdClientData, 
              argc, argv)
    TraceInfo_pt  TraceInfoP;
    Tcl_Interp   *interp;
    int           level;
    char         *command;
    int           (*cmdProc)();
    ClientData    cmdClientData;
    int           argc;
    char         *argv[];
{
        int i;

        printf("%2d", level);

	if (level > 20) level = 20;
        for (i = 0; i < level; i++) 
                printf("  ");

        if (TraceInfoP->NoEval) {
                if (TraceInfoP->NoTruncate)
                        printf("%s",command);
                else
                        {
                        printf("%.60s",command);
                        if (strlen(command) > CMD_TRUNCATE_SIZE)
                                printf("...");
                        }
        } else {
                for (i = 0; i < argc; i++) 
                        PrintArg (argv[i], TraceInfoP->NoTruncate);
        }

        putchar('\n');
        return;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdTrace --
 *     Implements the TCL trace command:
 *         trace level [noeval] [notruncate]
 *         trace off
 *         trace depth
 *     See the trace(TCL) manual page.
 *
 * Results:
 *      Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
int
CmdTrace (TraceInfoP, interp, argc, argv)
    TraceInfo_pt  TraceInfoP;
    Tcl_Interp   *interp;
    int           argc;
    char        **argv;
{
        int i;

        if ((argc < 2) || (argc > 4))
                return Tcl_RetError (interp, "TCL", 
                  "bad # arg: %.50s level [noeval] [notruncate] | off | depth",
                  argv[0]);

        TraceInfoP->NoEval = FALSE;
        TraceInfoP->NoTruncate = FALSE;
        for (i = 2; i < argc; i++) {
                if (strcmp(argv[i],"notruncate") == 0)
                        TraceInfoP->NoTruncate = TRUE;
                else if (strcmp(argv[2],"noeval") == 0)
                        TraceInfoP->NoEval = TRUE;
                else
                        return Tcl_RetError (interp, "TCL", 
                                 "noeval or notruncate option expected");
        }

        if (strcmp (argv[1],"depth") == 0) {
                sprintf(interp->result, "%d", TraceInfoP->Depth);
                return TCL_OK;
        }

        if (TraceInfoP->TraceHolder != NULL)
                {
                Tcl_DeleteTrace(interp, TraceInfoP->TraceHolder);
                TraceInfoP->Depth = 0;
                TraceInfoP->TraceHolder = NULL;
                }

        if (strcmp(argv[1],"off") == 0)
                return TCL_OK;

        if (!StrToInt (argv[1], 10, &(TraceInfoP->Depth)))
                return Tcl_RetError (interp, "TCL", 
                                     "Invalid numeric value for level");

        TraceInfoP->TraceHolder = 
                Tcl_CreateTrace (interp, TraceInfoP->Depth, TraceRoutine, 
                                 (ClientData)TraceInfoP);
        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 *  Tcl_InitDebug --
 *      Initialize the TCL debugging commands.
 *
 *----------------------------------------------------------------------
 */
Tcl_InitDebug (interp)
    Tcl_Interp *interp;
{
TraceInfo_pt InfoP;

InfoP = (TraceInfo_pt)ckalloc (sizeof (struct TraceInfo_t));
InfoP->TraceHolder = NULL;
InfoP->NoEval = FALSE;
InfoP->NoTruncate = FALSE;
InfoP->Depth = 0;

Tcl_CreateCommand (interp, "trace", CmdTrace, (ClientData)InfoP, 
                  (void (*)())NULL);
}


