#==============================================================================
# help.tcl --
#     Tcl help command.
#
#@help: help
#    help pathname
#    helpcd subjectdir
#
#    Help is a command used to display information about TCL commands, 
#    special variables and features.  Help files are structured in a 
#    multilevel directory tree of subjects and help files.  If you just
#    enter the command:
#        help
#
#    a listing of all top-level help subject directories and files is
#    displayed.  If you want to see all help files and lower level subject
#    directories under a directory, say:
#        help subjectdir
#
#    To display a help file, type:
#        help helpfile
#    or
#        help subjectdir/helpfile
# 
#    There is a concept of a current subject directory, much like the
#    unix current directory.  This directory can be changed with the
#    helpcd command.  To change the current subject directory, use:
#        helpcd subjectdir
#
#    To return to the top-level of the help tree, type:
#        helpcd /
#
#    Help path names may also include ".." elements. To display the 
#    current help subject directory, use:
#        helppwd
#
#    To print this description of help at any level, type:
#        help help
#    or
#        help ?
#
#@endhelp 
#==============================================================================

#------------------------------------------------------------------------------
# Take a path name which might have . and .. elements and flatten them out.

proc help:flattenPath {pathName} {
    set newPath ""
    foreach element [split / $pathName] {
        if [strequ $element "."] {
           continue
        }
        if [strequ $element ".."] {
            if {[length [join / $newPath]] == 0} {
                error "Help: name goes above subject directory root"}
            replace newPath [expr [length $newPath]-1]
            continue
        }
        strcat newPath " [list $element]"
    }
    set newPath [join / $newPath]
    
    # Take care of the case where we started with something line "/" or "/."

    if {[strequ $newPath ""] && [string match "/*" $pathName]} {
        set newPath "/"}
        
    return $newPath
}

#------------------------------------------------------------------------------
# Take the help current directory and a path and evaluate it into a help root-
# based path name.

proc help:EvalPath {pathName} {
    global helpGlob

    if {![string match "/*" $pathName]} {
        if [strequ $pathName ""] {
            return $helpGlob(curDir)}
        if [strequ $helpGlob(curDir) "/"] {
            set pathName /$pathName
        } else {
            set pathName $helpGlob(curDir)/$pathName}
    }
    set pathName [help:flattenPath $pathName]
    if {[string match "*/" $pathName] && ![strequ $pathName "/"]} {
        set pathName [substr $pathName 0 [expr [length $pathName]-1]]}

    return $pathName    
}

#------------------------------------------------------------------------------
# Display a line of output, pausing waiting for input before displaying if the
# screen size has been reached.

proc help:Display {line} {
    global helpGlob
    if {$helpGlob(lineCnt) >= 23} {
        puts stdout ":" no_nl
        flush stdout
        gets stdin discard
        set helpGlob(lineCnt) 0
    }
    puts stdout $line
    incr helpGlob(lineCnt)
}

#------------------------------------------------------------------------------
# Display a file.

proc help:DisplayFile {filepath} {

    set inFH [open $filepath r]
    while {![eof $inFH]} {
        gets $inFH fileBuf
        help:Display $fileBuf
    }
    close $inFH

}    

#------------------------------------------------------------------------------
# Procedure to return contents of a directory.  A list is returned, consisting
# of two lists.  The first list are all the directories (subjects) in the
# specified directory.  The second is all of the help files.  Eash sub-list
# is sorted in alphabetical order.
#

proc help:ListDir {dirPath} {
    set dirList ""
    set fileList ""
    if {[catch {set dirFiles [glob $dirPath/*]}] != 0} {
        error "No files in subject directory: $dirPath"}
    foreach fileName $dirFiles {
        if [file $fileName isdirectory] {
            strcat dirList "[file $fileName tail]/ "
        } else {
            strcat fileList "[file $fileName tail] "}
    }
   return [list [sortlist $dirList] [sortlist $fileList]]
}

#------------------------------------------------------------------------------
# Display a list of file names in a column format. This use columns of 14 
# characters 3 blanks.

proc help:DisplayColumns {nameList} {
    set count 0
    set outLine ""
    foreach name $nameList {
        if {$count == 0} {
            strcat outLine "   "}
        strcat outLine $name
        if {[incr count] < 4} {
            set padLen [expr 17-[clength $name]]
            if {$padLen < 3} {
               set padLen 3}
            strcat outLine [replicate " " $padLen]
        } else {
           help:Display $outLine
           set outLine ""
           set count 0
        }
    }
    if {$count != 0} {
        help:Display $outLine}
    return
}


#------------------------------------------------------------------------------
# Help command main.

proc help {args} {
    global helpGlob

if [empty $helpGlob(root)] {
	echo "'make bldhelp' must be done to install help before help will work"
	return
}

    if {[length $args] > 1} {
        error "Help: too many arugments"}

    set helpGlob(lineCnt) 0

    # Special case "help help", so we can get it at any level.

    if {[strequ $args help] || [strequ $args "?"]} {
        help:DisplayFile $helpGlob(root)/help
        return
    }

    set request [help:EvalPath $args]
    set requestPath $helpGlob(root)$request

    if {![file $requestPath exists]} {
        error "Help:\"$request\" does not exist"}
    
    if [file $requestPath isdirectory] {
        set dirList [help:ListDir $requestPath]
        set subList  [index $dirList 0]
        set fileList [index $dirList 1]
        if {[length $subList] != 0} {
            help:Display "\nSubjects available in $request:"
            help:DisplayColumns $subList
        }
        if {[length $fileList] != 0} {
            help:Display "\nHelp files available in $request:"
            help:DisplayColumns $fileList
        }
    } else {
        help:DisplayFile $requestPath
    }
    return
}


#------------------------------------------------------------------------------
# Helpcd main.
#   
# The name of the new current directory is assembled from the current 
# directory and the argument.  The name will be flatten and any trailing
# "/" will be removed, unless the name is just "/".

proc helpcd {args} {
    global helpGlob

    if {[length $args] == 0} {
        set args "/"
    } else {
        if {[length $args] > 1} {
            error "Helpcd: too many arugments"}
    }
    
    set request [help:EvalPath $args]
    set requestPath $helpGlob(root)$request

    if {![file $requestPath exists]} {
        error "Helpcd: \"$request\" does not exist"}
    
    if {![file $requestPath isdirectory]} {
        error "Helpcd: \"$request\" is not a directory"}

    set helpGlob(curDir) $request
    return    
}

#------------------------------------------------------------------------------
# Helpcd main.

proc helppwd {} {
        global helpGlob
        echo "Current help subject directory: $helpGlob(curDir)"
}
#------------------------------------------------------------------------------
# One time initialization done when the file is sourced.
#
global helpGlob TCL_PATHLIST

set helpGlob(root) [searchpath $TCL_PATHLIST help]
set helpGlob(curDir) "/"
set helpGlob(outBuf) ""

